/*
 * uninformed research
 * -------------------
 *
 * ELF manipulation library
 *
 * skape
 * mmiller@hick.org
 * 02/01/2003
 */
#ifndef _MELF_H
#define _MELF_H

#include <elf.h>

#define IS_OVERFLOW(melf, base, addr) ((melf)->imageSize < ((unsigned long)(addr) - (unsigned long)(base)))

typedef struct _elf_spec_header {

	unsigned long           index;
	unsigned long           identifier;

	union
	{

		Elf32_Shdr section;
		Elf32_Phdr program;

	} spec;

	void                    *content;
	unsigned long           contentLength;

	struct _elf_spec_header *prev;
	struct _elf_spec_header *next;

} ELF_SPEC_HEADER;

typedef struct _elf_spec_header_list {

	ELF_SPEC_HEADER *head;
	ELF_SPEC_HEADER *tail;

	unsigned long   seq;
	unsigned long   length;

} ELF_SPEC_HEADER_LIST;

typedef struct _melf {

	// Image itself
	char                 image[1024];
	unsigned long        imageSize;

	// Image header
	Elf32_Ehdr           header;

	// Section header table
	ELF_SPEC_HEADER_LIST sections;

	// Program header table
	ELF_SPEC_HEADER_LIST programs;

} MELF;

MELF *melf_new();
MELF *melf_open(const char *image);
void melf_destroy(MELF *melf);

unsigned long melf_save(MELF *melf, const char *path);

// Elf header functions

void melf_elfSetType(MELF *melf, Elf32_Half type);
Elf32_Half melf_elfGetType(MELF *melf);

void melf_elfSetMachine(MELF *melf, Elf32_Half machine);
Elf32_Half melf_elfGetMachine(MELF *melf);

void melf_elfSetVersion(MELF *melf, Elf32_Word version);
Elf32_Word melf_elfGetVersion(MELF *melf);

void melf_elfSetEntry(MELF *melf, Elf32_Addr entry);
Elf32_Addr melf_elfGetEntry(MELF *melf);

void melf_elfSetProgramHeaderOffset(MELF *melf, Elf32_Off offset);
Elf32_Off melf_elfGetProgramHeaderOffset(MELF *melf);

void melf_elfSetSectionHeaderOffset(MELF *melf, Elf32_Off offset);
Elf32_Off melf_elfGetSectionHeaderOffset(MELF *melf);

void melf_elfSetProgramHeaderCount(MELF *melf, Elf32_Half count);
Elf32_Half melf_elfGetProgramHeaderCount(MELF *melf);

void melf_elfSetSectionHeaderCount(MELF *melf, Elf32_Half count);
Elf32_Half melf_elfGetSectionHeaderCount(MELF *melf);

void melf_elfSetStringTableIndex(MELF *melf, Elf32_Half index);
Elf32_Half melf_elfGetStringTableIndex(MELF *melf);

// Section functions

ELF_SPEC_HEADER *melf_sectionAdd(MELF *melf);
ELF_SPEC_HEADER *melf_sectionGetEnum(MELF *melf);
ELF_SPEC_HEADER *melf_sectionEnumNext(MELF *melf, ELF_SPEC_HEADER *en);
ELF_SPEC_HEADER *melf_sectionEnumPrev(MELF *melf, ELF_SPEC_HEADER *en);

ELF_SPEC_HEADER *melf_sectionFindType(MELF *melf, unsigned long type);
ELF_SPEC_HEADER *melf_sectionFindIndex(MELF *melf, unsigned long index);
ELF_SPEC_HEADER *melf_sectionFindName(MELF *melf, const char *name);

void melf_sectionSetContent(MELF *melf, ELF_SPEC_HEADER *section, void *content, unsigned long contentLength);
void *melf_sectionGetContent(MELF *melf, ELF_SPEC_HEADER *section);

void melf_sectionSetName(MELF *melf, ELF_SPEC_HEADER *section, const char *name);
const char *melf_sectionGetName(MELF *melf, ELF_SPEC_HEADER *section);

void melf_sectionSetType(MELF *melf, ELF_SPEC_HEADER *section, unsigned long type);
unsigned long melf_sectionGetType(MELF *melf, ELF_SPEC_HEADER *section);

void melf_sectionSetAddress(MELF *melf, ELF_SPEC_HEADER *section, Elf32_Addr addr);
Elf32_Addr melf_sectionGetAddress(MELF *melf, ELF_SPEC_HEADER *section);

void melf_sectionSetFlags(MELF *melf, ELF_SPEC_HEADER *section, Elf32_Word flags);
Elf32_Word melf_sectionGetFlags(MELF *melf, ELF_SPEC_HEADER *section);

unsigned long melf_sectionGetIndex(MELF *melf, ELF_SPEC_HEADER *section);

void melf_sectionSetStringTableHeader(MELF *melf, ELF_SPEC_HEADER *section, ELF_SPEC_HEADER *stringTable);
ELF_SPEC_HEADER *melf_sectionGetStringTableHeader(MELF *melf, ELF_SPEC_HEADER *section);

unsigned long melf_sectionRemove(MELF *melf, unsigned long id);

// Program header functions

ELF_SPEC_HEADER *melf_programAdd(MELF *melf);
ELF_SPEC_HEADER *melf_programGetEnum(MELF *melf);
ELF_SPEC_HEADER *melf_programEnumNext(MELF *melf, ELF_SPEC_HEADER *en);
ELF_SPEC_HEADER *melf_programEnumPrev(MELF *melf, ELF_SPEC_HEADER *en);

void melf_programSetType(MELF *melf, ELF_SPEC_HEADER *program, Elf32_Word type);
Elf32_Word melf_programGetType(MELF *melf, ELF_SPEC_HEADER *program);

void melf_programSetVirtualAddress(MELF *melf, ELF_SPEC_HEADER *program, Elf32_Addr addr);
Elf32_Addr melf_programGetVirtualAddress(MELF *melf, ELF_SPEC_HEADER *program);

void melf_programSetPhysicalAddress(MELF *melf, ELF_SPEC_HEADER *program, Elf32_Addr addr);
Elf32_Addr melf_programGetPhysicalAddress(MELF *melf, ELF_SPEC_HEADER *program);

void melf_programSetVirtualSize(MELF *melf, ELF_SPEC_HEADER *program, Elf32_Word size);
Elf32_Word melf_programGetVirtualSize(MELF *melf, ELF_SPEC_HEADER *program);

void melf_programSetPhysicalSize(MELF *melf, ELF_SPEC_HEADER *program, Elf32_Word size);
Elf32_Word melf_programGetPhysicalSize(MELF *melf, ELF_SPEC_HEADER *program);

void melf_programSetFlags(MELF *melf, ELF_SPEC_HEADER *program, Elf32_Word flags);
Elf32_Word melf_programGetFlags(MELF *melf, ELF_SPEC_HEADER *program);

unsigned long melf_programRemove(MELF *melf, unsigned long id);

// String table manipulation

unsigned long melf_stringTableSetString(MELF *melf, ELF_SPEC_HEADER *stringTable, const char *name);
const char *melf_stringTableGetString(MELF *melf, ELF_SPEC_HEADER *stringTable, unsigned long index);

// Dynamic manipulation

ELF_SPEC_HEADER *melf_dynamicCreate(MELF *melf);

Elf32_Dyn *melf_dynamicAddTag(MELF *melf, ELF_SPEC_HEADER *dynamic, Elf32_Sword tag, Elf32_Word val);
unsigned long melf_dynamicRemoveTag(MELF *melf, ELF_SPEC_HEADER *dynamic, Elf32_Sword tag);
unsigned long melf_dynamicRemoveIndex(MELF *melf, ELF_SPEC_HEADER *dynamic, unsigned long index);
Elf32_Dyn *melf_dynamicGetTag(MELF *melf, ELF_SPEC_HEADER *dynamic, Elf32_Sword tag);
Elf32_Dyn *melf_dynamicGetIndex(MELF *melf, ELF_SPEC_HEADER *dynamic, unsigned long index);
unsigned long melf_dynamicSetTag(MELF *melf, ELF_SPEC_HEADER *dynamic, Elf32_Sword tag, Elf32_Word val);
unsigned long melf_dynamicSetIndex(MELF *melf, ELF_SPEC_HEADER *dynamic, unsigned long index, Elf32_Word val);

// Note manipulation

ELF_SPEC_HEADER *melf_noteCreate(MELF *melf, const char *name, unsigned long createProgramHeader);

Elf32_Nhdr *melf_noteAdd(MELF *melf, ELF_SPEC_HEADER *note, Elf32_Word type, const char *name, unsigned char *desc, unsigned long descLength);
Elf32_Nhdr *melf_noteEnum(MELF *melf, ELF_SPEC_HEADER *note, unsigned long index);
unsigned long melf_noteRemove(MELF *melf, ELF_SPEC_HEADER *note, Elf32_Nhdr *item);

unsigned long melf_noteGetType(MELF *melf, ELF_SPEC_HEADER *note, Elf32_Nhdr *item);
const char *melf_noteGetName(MELF *melf, ELF_SPEC_HEADER *note, Elf32_Nhdr *item);
unsigned char *melf_noteGetDesc(MELF *melf, ELF_SPEC_HEADER *note, Elf32_Nhdr *item);

#ifdef MELF_COMP
#include "melf_internal.h"
#endif

#endif
